#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>

/* Standard includes for OSS DSP using */
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/soundcard.h>

#include "ayemu.h"

static const int DEBUG = 0;

struct option options[] = {
  { "quiet",   no_argument, NULL, 'q'},
  { "rand",    no_argument, NULL, 'Z'},
  { "stdout",  no_argument, NULL, 's'},
  { "verbose", no_argument, NULL, 'v'},
  { "usage",   no_argument, NULL, 'u'},
  { "help",    no_argument, NULL, 'h'},
  { 0, 0, 0, 0}
};

ayemu_ay_t ay;
ayemu_vtx_t vtx;
char regs[14];
uint8_t * audio_buf;
int audio_bufsize;
int audio_fd;
int freq, chans, bits;

#define DEVICE_NAME "/dev/dsp"


void usage ()
{
    printf(
	   "AY/YM VTX format player.\n"
	   "It uses libayemu (http://libayemu.sourceforge.net)\n"
	   "Version 0.1 (2004/Oct/27). Written and copyrights by Alexander Sashnov.\n"
	   "THIS SOFTWARE COMES WITH ABSOLUTELY NO WARRANTY! USE AT YOUR OWN RISK!\n\n"
	   "Usage: playvtx [option(s)] files...\n"
	   "  -q --quiet\tquiet (don't print title)\n"
	   "  -Z --random\trandom play\n"
	   "  -s --stdout\twrite to stdout\n"
	   "  -v --verbose\tincrease verbosity level\n"
	   "  -h --help\n"
	   "  -u --usage\tthis help\n"
	   );
}

/* play vtx file */
void Play (const char *filename);

int
     main (int argc, char **argv)
     {
       int qflag = 0;  // quite plating
       int Zflag = 0;  // random list
       int sflag = 0;  // to stdout
       int vflag = 0;  // verbose
       int index;
       int c;
       /* `getopt_long' stores the option index here. */
       int option_index = 0;
	   
       opterr = 0;
     
       while ((c = getopt_long (argc, argv, "+qZsvuh", options, &option_index)) != -1)
         switch (c)
           {
           case 'q':
             qflag = 1;
             break;
           case 'Z':
             Zflag = 1;
             break;
           case 's':
             sflag = 1;
             break;
           case 'v':
             vflag = 1;
             break;
	   case 'h':
	   case 'u':
	     usage ();
	     exit (0);
           case '?':
             if (isprint (optopt))
               fprintf (stderr, "Unknown option `-%c'.\n", optopt);
	     else
               fprintf (stderr,
                        "Unknown option character `\\x%x'.\n",
                        optopt);
	     usage ();
             return 1;
	   case -1:
	     break;
           default:
             abort ();
           }
     
       if (DEBUG)
       printf ("qflag = %d, Zflag = %d, sflag = %d, vflag = %d\n", 
	       qflag, Zflag, sflag, vflag);
     
       if (Zflag)
	 printf ("The -Z flag is not implemented yet, sorry\n");


       for (index = optind; index < argc; index++)
	 {
	 if (DEBUG) 
	   printf ("Playing file %s\n", argv[index]);

	 Play (argv[index]);
	 }
       return 0;
     }


void Play (const char *filename)
{
  int n, len;

  memset (&ay, 0, sizeof(ay));

  freq = 44100;
  chans = 2;  /* 1=mono, 2=stereo */
  bits = 16;  /* 16 or 8 bit */

  /* Try to load VTX file */
  if (! ayemu_vtx_open (&vtx, filename)) {
    fprintf (stderr, "Can't open vtx file %s\n", filename);
  }
  else if (! ayemu_vtx_load_data (&vtx)) {
    fprintf (stderr, "Can't unpack vtx data");
    goto free_vtx;
  }
    /* =====----- OSS Init sound system ------======= */
  else if ((audio_fd = open(DEVICE_NAME, O_WRONLY, 0)) == -1) {
    fprintf (stderr, "Can't open /dev/dsp\n");
  }
  else if (ioctl(audio_fd, SNDCTL_DSP_SETFMT, &bits) == -1) {
    fprintf (stderr, "Can't set sound format\n");
  }
  else if (ioctl(audio_fd, SNDCTL_DSP_CHANNELS, &chans) == -1) {
    fprintf (stderr, "Can't set number of channels\n");
  }
  else if (ioctl(audio_fd, SNDCTL_DSP_SPEED, &freq) == -1) {
    fprintf (stderr, "Can't set audio freq\n");
  }
  else {

    /* VTX song info */
    printf ("Title: %s\n", &vtx.Title);
    printf ("Author: %s\n", &vtx.Author);
    printf ("From : %s\n", &vtx.From);
    printf ("Tracker : %s\n", &vtx.Tracker);
    printf ("Comment : %s\n", &vtx.Comment);
    printf ("Chip type: %s\tStereo: %d\tLoop: %d\n", vtx.Chip, vtx.Stereo, vtx.Loop);
    printf ("Chip freq: %d\tPlayer Freq:%d\tYear: %d\n", vtx.ChipFreq, vtx.PlayerFreq, vtx.Year);

    if (DEBUG) printf ("OSS sound system initialization success: bits=%d, chans=%d, freq=%d\n", bits, chans, freq);

    /* Allocate audio buffer for one AY frame */
    int audio_bufsize = freq * chans * (bits >> 3) / vtx.PlayerFreq;
    if ((audio_buf = (uint8_t *) malloc (audio_bufsize)) == NULL) {
      fprintf (stderr, "Can't allocate sound buffer\n");
      goto free_audio;
    }

    ayemu_set_chip (&ay, (strcmp(vtx.Chip,"ay") ==0)? AYEMU_AY:AYEMU_YM,
        vtx.ChipFreq, (ayemu_stereo_t)vtx.Stereo);
    if (DEBUG) printf ("ayemu_set_chip (%2s, freq %d, stereo %d)\n",vtx.Chip, vtx.ChipFreq, vtx.Stereo);

    ayemu_start (&ay, freq, chans, bits);
    if (DEBUG) printf ("ayemu_start (freq %d, chans %d, bits %d)\n",freq,chans,bits);

    while ( ayemu_vtx_get_next_frame (&vtx, (void*)&regs)) {
      ayemu_set_regs (&ay, (void*)&regs);
      if (DEBUG)
	printf ("VTX pos %4d Regs: tones: %4d %4d %4d %2d R7: %d%d%d %d%d%d vols: %2d %2d %2d env: %5d %2d\n", vtx.pos,
		ay.reg_tone_a, ay.reg_tone_b, ay.reg_tone_c, ay.reg_noise, 
		ay.reg_R7_tone_a, ay.reg_R7_tone_b, ay.reg_R7_tone_c, 
		ay.reg_R7_noise_a, ay.reg_R7_noise_b, ay.reg_R7_noise_c,
		ay.reg_vol_a, ay.reg_vol_b, ay.reg_vol_c,
		ay.reg_env_freq, ay.reg_env_style);

      ayemu_gen_sound (&ay, audio_buf, audio_bufsize);

      /* play sound buffer */
      if ((len = write(audio_fd, audio_buf, audio_bufsize)) == -1) {
	fprintf (stderr, "Error writting to sound device, break.\n");
	break;
      }
    }

  free_audio:
    close (audio_fd);
  free_ay:   
    ayemu_free(&ay); 
  free_vtx:
    ayemu_vtx_free(&vtx);
  }
}
